# C++ 浅谈Ring Buffer
# https://mp.weixin.qq.com/s/z2JzgS8dt04SJgWPT1v24w
import os
import shutil
import time
from pathlib import Path
from typing import Optional, Union

import numpy as np
from loguru import logger
from typing_extensions import Literal  # 3.8+
from typing_extensions import Self  # 3.11+

from npyt.format import get_file_ctx, save, load, resize


class NPYT:
    """
    带尾巴的NPY格式文件,后面带小尾巴，为数据开始和结束位置

    - raw: 原始数据区。与内存映射文件循序一致，为最大长度。可以修改。
    - data: 有效数据区。为有效数据区的长度。环形会拼接成整块。可以修改。

    Notes
    -----
    很多操作没有条件判断，使用时需要注意。

    """

    def __init__(self, filename: Union[str, Path], dtype: Optional[np.dtype] = None):
        """初始化

        Parameters
        ----------
        filename:str
            文件名
        dtype:np.dtype
            数据类型。
                to_records时，数据类型可能与预期的不一致。append时可能取的dtyep不满足。所以在这提前限制更合适

        """
        self._filename: Path = Path(filename)
        self._t: Optional[np.ndarray] = None
        self._a: Optional[np.ndarray] = None
        self._capacity: int = 0
        self._tell: int = 0
        self._dtype: Optional[np.dtype] = dtype

    def filename(self) -> Path:
        return self._filename

    def _test(self, start: int, end: int) -> None:
        """测试用。强行设置头尾指针"""
        self._t[0] = start
        self._t[1] = end

    def _raw(self) -> np.ndarray:
        """测试用。取原始数组"""
        return self._a

    def _raw_len(self) -> int:
        """测试用。获取原始数组长度"""
        return self._a.shape[0]

    def info(self):
        """获取头信息"""
        """获取尾巴关键信息"""
        if self._t is None:
            return None
        return tuple(self._t.tolist())

    def dtype(self) -> np.dtype:
        """获取数据类型"""
        return self._dtype

    def clear(self) -> Self:
        """重置位置指针，相当于清空了数据"""
        self._t[0:2] = 0
        self._t[2] = time.time_ns()
        return self

    def start(self) -> int:
        """获取缓冲区开始位置"""
        # return int(self._t[0])
        return 0

    def end(self) -> int:
        """获取缓冲区结束位置"""
        return int(self._t[1])

    def timestamp(self) -> int:
        """获取时间戳"""
        return int(self._t[2])

    def empty(self) -> bool:
        """查询缓冲区是否为空"""
        return self.start() == self.end()

    def full(self) -> bool:
        """查询缓冲区是否已满"""
        return (self.end() + 1) % (self._capacity + 1) == self.start()

    def size(self) -> int:
        """当前缓冲区中元素个数"""
        start, end = self.start(), self.end()
        if end >= start:
            return end - start
        else:
            return self._capacity - (start - end)

    def capacity(self) -> int:
        """缓冲区容量大小（最大可容纳元素数）"""
        return self._capacity

    def load(self, mmap_mode: Literal["r", "r+"]) -> Self:
        """加载文件。为以后操作做准备

        Parameters
        ----------
        mmap_mode
            内存文件映射模式

            r: 只读
            r+: 读写

        """
        self._a, self._t = load(self._filename, mmap_mode=mmap_mode)
        self._capacity = self._a.shape[0]
        if self._dtype is None:
            self._dtype = self._a.dtype
        else:
            assert self._dtype == self._a.dtype, f"dtype mismatch {self._dtype} != {self._a.dtype}"
        return self

    def save(self,
             array: Optional[np.ndarray] = None,
             capacity: int = 0,
             end: Optional[int] = None,
             skip_if_exists: bool = True) -> Self:
        """创建文件

        Parameters
        ----------
        array: np.ndarry
            初始数据.如果为None，将创建一个空数组。
        capacity:int
            最大容量
        end:int
            结束位置。0表示只创建文件，数据区为空。None表示使用array的长度。
        skip_if_exists:bool
            如果文件已经存在了就跳过。反之新建

        """
        if skip_if_exists and self._filename.exists():
            return self

        if array is None:
            array = np.empty((1,), dtype=self._dtype)
            end = 0
        elif self._dtype is None:
            self._dtype = array.dtype
        else:
            assert self._dtype == array.dtype, f"dtype mismatch {self._dtype} != {array.dtype}"
        save(get_file_ctx(self._filename, mode="wb+"), array, capacity, end)

        return self

    def resize(self, capacity: Optional[int] = None) -> bool:
        """文件截断或扩充。不能丢失有效数据

        Parameters
        ----------
        capacity:int
            新的长度。

            None将截断数据到有效长度

        Notes
        -----
        独占时才能使用

        """
        # 一些基本信息
        start, end = self.start(), self.end()

        # 数据环形，文件不能动了
        if end < start:
            return False

        # 有效右边界
        if capacity is None:
            capacity = end
        capacity = max(end, capacity)

        # 一定要copy,因为后面要释放文件
        arr = self._a[:1].copy()
        # 释放文件占用
        self._a = None
        self._t = None
        # 释放后就可以动文件了
        return resize(self._filename, arr, start, end, capacity)

    def backup(self, to_path: Union[str, Path]) -> None:
        """备份

        Parameters
        ----------
        to_path:str
            备份路径

        """
        path = Path(to_path)
        path.mkdir(parents=True, exist_ok=True)

        shutil.copy2(self._filename, path)
        logger.info("backup {} to {}", self._filename, path)

    def data(self) -> np.ndarray:
        """取数据区。环形数据会拼接起来不可修改"""
        start, end = self.start(), self.end()
        return self._a[start:end]

    def head(self, n: int = 5) -> np.ndarray:
        """取头部数据"""
        start, end = self.start(), self.end()
        return self._a[start:min(start + n, end)]

    def tail(self, n: int = 5) -> np.ndarray:
        """取尾部数据"""
        start, end = self.start(), self.end()
        return self._a[max(start, end - n):end]

    def at(self, index) -> np.ndarray:
        """取某一行数据

        Parameters
        ----------
        index:int
            索引

        """
        return self._a[index]

    def append(self, array: np.ndarray) -> int:
        """缓冲区插入函数

        Parameters
        ----------
        array:
            插入的数据

        Returns
        -------
        int
            剩余未插入的行数

        Notes
        -----
        1. 插入单行数据时要注意，shape为(1, n)，而不是(n,)
            - arr[0:1] 正确
            - arr[0] 错误

        2. a[0:0] = b[0:1] 不报错，也没保存
            (0, 3)  (1, 3)

        3. a[10:20] = b[0:1] 不报错，也没保存
            (0, 3)  (1, 3)

        """
        remaining = array.shape[0]
        # 空内容，没必要
        if remaining == 0:
            return remaining

        end = self.end()
        _end = end + remaining
        # 空间不够，直接返回剩余行数
        if _end > self._raw_len():
            return remaining

        self._a[end:_end] = array
        self._t[1] = _end
        self._t[2] = time.time_ns()

        return 0

    def expend(self, array: np.ndarray) -> bool:
        """缓冲区插入函数，空间不够扩充文件大小

        Parameters
        ----------
        array:
            插入的数据

        Returns
        -------
        int
            剩余未插入的行数

        See Also
        --------
        append: 空间不够直接返回剩余行数

        """
        remaining = array.shape[0]
        # 空内容，没必要
        if remaining == 0:
            return True

        end = self.end()
        _end = end + remaining
        # 数据空了，后面空间也不够，移动到开头
        if _end > self._raw_len():
            if self.resize(_end):
                self.load("r+").append(array)
                return True
            else:
                self.load("r+")
                return False

        self._a[end:_end] = array
        self._t[1] = _end
        self._t[2] = time.time_ns()

        return True

    def remove(self) -> bool:
        """删除文件"""
        self._a = None
        self._t = None
        try:
            os.remove(self._filename)
            logger.trace("remove {}", self._filename.resolve())
            return True
        except PermissionError as e:
            logger.error("remove {} error:{}", self._filename.resolve(), e)
            return False

    def merge(self, obj: Self) -> bool:
        """合并文件"""
        if self.expend(obj.data()):
            return obj.remove()
        return False

    def rename(self, name) -> bool:
        """重命名。如果文件已经存在了会被覆盖"""
        self._a = None
        self._t = None
        shutil.move(self._filename, name)
        self._filename = Path(name)
        return True

    def tell(self) -> int:
        return self._tell

    def rewind(self) -> Self:
        """重置当前指针到数据的起始位置

        Notes
        -----
        `pop`后导致start发生变化，根据业务需要，可能要重置指针到start

        """
        self._tell = self.start()
        return self

    def seek(self, offset: int, whence: int = 0) -> Self:
        """在start:end范围内seek

        Parameters
        ----------
        offset:int
            >0: 向后偏
            <0: 向前偏
        whence:int
            0,1,2

        """
        start, end = self.start(), self.end()
        if whence == os.SEEK_SET:
            _curr = start
        elif whence == os.SEEK_CUR:
            _curr = self._tell
        elif whence == os.SEEK_END:
            _curr = end
        else:
            _curr = end

        self._tell = max(min(_curr + offset, end), start)
        return self

    def read(self, n: int = 1024, prefetch: int = 0) -> np.ndarray:
        """读取n行数据。不移动start指针，而是移动tell指针

        Parameters
        ----------
        n:int
            读取行数
        prefetch:int
            预读取行数。需>=0

        Returns
        -------
        np.ndarray
            读取的数据

        """
        start, end = self.start(), self.end()

        _start = max(self._tell - prefetch, start)
        self._tell = min(self._tell + n, end)

        arr = self._a[_start:self._tell]

        return arr



import contextlib
import os
import time
from pathlib import Path
from typing import Optional, Literal, Tuple

import numpy as np
from loguru import logger
from numpy.lib.format import GROWTH_AXIS_MAX_DIGITS  # noqa
from numpy.lib.format import _write_array_header  # noqa
from numpy.lib.format import dtype_to_descr

"""
添加的小尾巴，欢迎提供更好的格式方案

1. start: 开始位置
2. end: 结束位置
3. update_timestamp: 修改时间，用来判读是否要读取，用来解决end不变的情况
4. offset: 数据区开始位置，方便其他语言快速定位并写入
5. magic: 魔术数，用来判断是否`NPYT`格式文件

"""
_TAIL_SIZE_: int = 5
_TAIL_ITEMSIZE_: int = np.dtype(np.uint64).itemsize * _TAIL_SIZE_
_MAGIC_NUMBER_: int = 20250510_080000  # 2025年05月10日 东八区

# 记录原始函数
_origin_descr_to_dtype = np.lib.format.descr_to_dtype


def dtype_with_align(dtype: np.dtype):
    """想办法为dtype加上align参数

    np.load加载的dtype.descr不含alignment，表面上可以正常使用，但与numba配合时报错
    float64位转float32报错，因为类型不同导致了复制，然后复制的是只读不能写入

    """
    # https://github.com/numpy/numpy/pull/24685
    if dtype.names is None:
        return dtype
    align = len(dtype.descr) > len(dtype.names)
    descr = [(x, y) for x, y in dtype.descr if x != '']
    return np.dtype(descr, align=align)


def _descr_to_dtype(descr):
    return dtype_with_align(_origin_descr_to_dtype(descr))


# 函数替换
np.lib.format.descr_to_dtype = _descr_to_dtype


def dtype_to_column_dtypes(dtype):
    """df.to_records的dtype转column_dtypes"""
    return {x: y for x, y in dtype.descr if x != ''}


class TuplePad(tuple):

    def __repr__(self):
        """npy头信息当想扩充或截断时shape字符串长度固定，提前占位"""
        items = ', '.join(f'{repr(x):>21}' if i == 0 else repr(x) for i, x in enumerate(self))
        if len(self) > 1:
            return f"({items})"
        if len(self) == 1:
            return f"({items},)"  # 只有一个元素时，末尾加逗号
        return f"()"  # 空元组


def get_shape(shape: tuple, capacity: int) -> tuple:
    """返回偏大的形状"""
    shape = list(shape)
    if len(shape) == 0:
        shape.append(0)
    # shape[0] = max(int(capacity + 1), shape[0] + 1)
    shape[0] = max(int(capacity), shape[0])
    shape = TuplePad(shape)
    return shape


def get_end(val, end: Optional[int] = None) -> int:
    """数据区结束位置。end为None时，返回val"""
    if end is None:
        end = val
    return int(end)


def get_nbytes(dtype: np.dtype, shape: tuple, offset: int) -> int:
    """得到数据区大小"""
    shape = tuple(shape)
    size = np.intp(1)
    for k in shape:
        size *= k

    return int(offset + size * dtype.itemsize)


def get_file_ctx(file, mode: str = "wb"):
    """文件上下文

    wb: 清空只写
    wb+: 清空读写

    """
    if hasattr(file, 'write'):
        file_ctx = contextlib.nullcontext(file)
    else:
        file = os.fspath(file)
        if not file.endswith('.npy'):
            file = file + '.npy'
        file_ctx = open(file, mode)
    return file_ctx


def write_header(fp, array: np.ndarray, shape: tuple) -> int:
    """写入头"""

    def header_data_from_array_1_0(array):
        d = {'shape': shape}

        if array.flags.c_contiguous:
            d['fortran_order'] = False
        elif array.flags.f_contiguous:
            d['fortran_order'] = True
        else:
            d['fortran_order'] = False

        d['descr'] = dtype_to_descr(array.dtype)
        return d

    _write_array_header(fp, header_data_from_array_1_0(array))
    return fp.tell()


def write_footer(fp, dtype: np.dtype, shape: tuple, start: int, end: int, offset: int) -> None:
    """定位 文件头+数据区，然后写入尾巴"""
    # 扩充文件大小
    fp.seek(get_nbytes(dtype, shape, offset), 0)
    # 写入尾巴
    fp.write(np.array([int(start), int(end), time.time_ns(), offset, _MAGIC_NUMBER_], dtype=np.uint64).tobytes())


def load(filename, mmap_mode: Literal["r", "r+", "w+"]) -> Tuple[np.ndarray, np.ndarray]:
    """加载带尾巴的NPY格式文件"""
    # dtype缺align，提前修改了函数np.lib.format.descr_to_dtype
    offset = os.path.getsize(filename)
    assert offset > _TAIL_ITEMSIZE_, f"文件大小不合法，非有效`NPYT`格式文件"
    tail = np.memmap(filename, shape=(_TAIL_SIZE_,), dtype=np.uint64, mode=mmap_mode,
                     offset=offset - _TAIL_ITEMSIZE_)
    if tail[4] != _MAGIC_NUMBER_:
        logger.warning(f"文件格式错误，不是`NPYT`格式文件，涉及到尾部信息的函数都不正确，谨慎使用")
        # 设置成None防止array被修改
        tail = None
    arr = np.load(filename, mmap_mode=mmap_mode)

    return arr, tail


def save(file_ctx, array: np.ndarray, capacity: int, end: Optional[int] = None) -> None:
    """保存"""
    shape = get_shape(array.shape, capacity)
    end = get_end(array.shape[0], end)

    with file_ctx as fp:
        # 写入头信息
        offset = write_header(fp, array, shape)
        if end > 0:
            # 写入数据
            array.tofile(fp)
        # 写入尾巴。带文件大小调整能力
        write_footer(fp, array.dtype, shape, 0, end, offset)
        fp.flush()


def resize(filename: Path, row: np.ndarray, start: int, end: int, capacity: Optional[int] = None) -> bool:
    """文件截断或扩充

    Parameters
    ----------
    filename:str
        文件名
    row:
        初始数据，只是提取dtype等使用。一般只取了一行。
    start:int
        开始位置
    end:int
        结束位置
    capacity:Optional[int]
        容量

    Notes
    -----
    独占时才能使用

    """
    # 其实是直接用的capacity生成shape
    shape = get_shape(row.shape, capacity)

    try:
        with get_file_ctx(filename, mode="r+b") as fp:
            old_size = fp.seek(0, 2)
            fp.seek(0, 0)
            offset = write_header(fp, row, shape)
            fp.seek(get_nbytes(row.dtype, shape, 0), 1)
            write_footer(fp, row.dtype, shape, start, end, offset)
            new_size = fp.tell()
            fp.truncate(new_size)
            fp.flush()
            logger.trace("resize {} from {} to {}", filename.resolve(), old_size, new_size)
        return True
    except PermissionError as e:
        logger.error("resize {} error:{}", filename.resolve(), e)
        return False



import shutil
import time
from pathlib import Path
from typing import List, Optional, Union

import more_itertools
import numpy as np
from loguru import logger
from typing_extensions import Self

from npyt import NPYT


class NPY8:

    def __init__(self, name: Union[str, Path], capacity_per_file: int = 1024, query_size: int = 8, dtype: Optional[np.dtype] = None):
        """无尽头增长文件

        Parameters
        ----------
        name:str
            目录名
        capacity_per_file:int
            每个子文件最大容量
        query_size:int
            队列长度

        """
        self._name: str = name
        self._capacity_per_file: int = capacity_per_file
        self._size: int = max(query_size, 2)
        self._dtype: Optional[np.dtype] = dtype

        self._path: Path = Path(self._name)
        self._path.mkdir(parents=True, exist_ok=True)
        self._path_lock = self._path / '.lock'
        # lock文件，维护了多个时间戳文件名
        self._lock: Optional[np.ndarray] = None
        self._writer: Optional[NPYT] = None
        self._reader: Optional[NPYT] = None
        # 正在读的文件名时间戳
        self._reader_ts: int = -1

    def capacity(self) -> int:
        """总容量。只是队列中的文件容量之和。与NPYT的接口保持相同"""
        return self._capacity_per_file * self._size

    def remove(self):
        """删除文件"""
        self._lock = None
        self._writer = None
        self._reader = None
        self._reader_ts = -1

        shutil.rmtree(self._path)
        logger.info("remove {} ", self._path.resolve())

    def load(self) -> Self:
        """初始化并加载"""
        self._lock = None
        self._writer = None
        self._reader = None
        self._reader_ts = -1

        if self._path_lock.exists():
            self._lock = np.memmap(self._path_lock, dtype=np.uint64, mode="r+", shape=(self._size,))
        else:
            self._lock = np.memmap(self._path_lock, dtype=np.uint64, mode="w+", shape=(self._size,))

        # 没有数据，初始化
        t = self._lock[0]
        filename = self._path / f'{t}.npy'
        if not filename.exists():
            files = sorted(self._path.glob('*.npy'))[-self._size:]
            self._lock[:] = 0
            for i, f in enumerate(files):
                self._lock[i] = int(f.stem)

        return self

    def append(self, data: np.ndarray) -> int:
        """添加数据，遇到文件空间不足时会新增文件

        Parameters
        ----------
        data:np.ndarray
            新数据

        Returns
        -------
        int
            成功返回0，失败返回剩余行数。由于会一直新增文件，理论上一直返回0

        """
        if self._writer:
            remaining = self._writer.append(data)
            if remaining == 0:
                return 0

            # 失败
            if self._lock[-1] > 0:
                # 队列满了，平移队列
                t = self._lock[0]
                self._lock[:-1] = self._lock[1:]
                self._lock[-1] = time.time_ns()
                # 出队列后修改文件大小。留心文件被占用导致失败
                filename = self._path / f'{t}.npy'
                if t == self._reader_ts:
                    # !!! 已经要出队列了，读指针还没切换，只能说是read调用频率太低，或队列太短
                    self._reader = None
                    logger.warning('{} is still reading. read too less or queue too short', filename.resolve())
                NPYT(filename).load(mmap_mode='r').resize()
            else:
                # 到下一个位置
                self._lock[np.argmax(self._lock) + 1] = time.time_ns()

        # 这样基本不会有空文件
        if self._lock[0] == 0:
            self._lock[0] = time.time_ns()

        # 找到最大编号文件。但不知道文件是否满了
        t = self._lock[np.argmax(self._lock)]
        filename = self._path / f'{t}.npy'
        if filename.exists():
            # 加载已有文件
            self._writer = NPYT(filename, dtype=self._dtype).load(mmap_mode="r+")
            return self.append(data)
        else:
            logger.trace("create {}", filename.resolve())
            # 可以一次性保存大文件
            self._writer = NPYT(filename, dtype=self._dtype).save(array=data, capacity=self._capacity_per_file).load(mmap_mode="r+")
            return 0

    def read(self, n: int = 1024, prefetch: int = 0) -> np.ndarray:
        """读取数据

        Parameters
        ----------
        n:int
            从当前位置，读取行数
        prefetch:int
            预加载行数

        Notes
        -----
        一次性取数限制在单个文件，当前文件读完后，再次读取自动切换成下一文件

        """
        if self._reader:
            arr = self._reader.read(n, prefetch)
            if len(arr) > 0:
                return arr

            # 未取得数据，又是最后一个文件，不切换文件，因为下次调用可能又有新数据了
            if np.max(self._lock) == self._reader_ts:
                return np.empty(0, dtype=self._dtype)

        # 未打开文件，或未取到数据到此
        condition = self._lock > self._reader_ts
        if not np.any(condition):
            # 没找到任何符合条件的文件. 这里永远都到不了？
            return np.empty(0, dtype=self._dtype)

        # 定位到第一个文件，或上次文件的下一个
        max_idx = np.argmax(condition)
        # 找到大于指针位置的文件。0也没关系，反正文件不存在
        t = self._lock[max_idx]
        filename = self._path / f'{t}.npy'
        self._reader_ts = t
        if filename.exists():
            # 加载已有文件
            self._reader = NPYT(filename).load(mmap_mode="r")
            return self.read(n, prefetch)
        else:
            # 文件没了，返回空
            return np.empty(0, dtype=self._dtype)

    def tail(self, n: int = 5) -> List[np.ndarray]:
        """取尾部数据

        Parameters
        ----------
        n:int
            行数

        Returns
        -------
        List[np.ndarray]
            一个文件对应一个np.ndarray

        """
        max_idx = np.argmax(self._lock)
        outputs = []
        remaining = n
        for i in range(max_idx, -1, -1):
            t = self._lock[i]
            filename = self._path / f'{t}.npy'
            if filename.exists():
                arr = NPYT(filename).load(mmap_mode="r").tail(remaining)
                outputs.insert(0, arr)
                remaining -= len(arr)
                if remaining <= 0:
                    break

        return outputs

    def head(self, n: int = 5) -> List[np.ndarray]:
        """取头部数据

        Parameters
        ----------
        n:int
            行数

        Returns
        -------
        List[np.ndarray]
            一个文件对应一个np.ndarray

        """
        max_idx = np.argmax(self._lock)
        outputs = []
        remaining = n
        for i in range(0, max_idx):
            t = self._lock[i]
            filename = self._path / f'{t}.npy'
            if filename.exists():
                arr = NPYT(filename).load(mmap_mode="r").head(remaining)
                outputs.append(arr)
                remaining -= len(arr)
                if remaining <= 0:
                    break

        return outputs

    def start(self) -> int:
        return 0

    def end(self) -> int:
        """当前写入位置.如果没有打开文件，尝试打开，失败返回0"""
        if self._writer:
            return self._writer.end()

        max_idx = np.argmax(self._lock)
        for i in range(max_idx, -1, -1):
            t = self._lock[i]
            filename = self._path / f'{t}.npy'
            if filename.exists():
                self._writer = NPYT(filename, dtype=self._dtype).load(mmap_mode="r+")
                return self._writer.end()

        return 0

    def merge(self, batch_size: int = 4) -> bool:
        """合并文件，并改名

        Parameters
        ----------
        batch_size:int
            每批次大小

        Notes
        -----
        合并的是不在队列中维护的文件。不影响当前读写指针

        """
        batch_size = max(batch_size, 2)
        files = sorted(self._path.glob('*.npy'))[:-self._size]
        for batch in more_itertools.batched(files, batch_size):
            if len(batch) != batch_size:
                continue
            f1 = NPYT(batch[0]).load(mmap_mode="r+")
            for f in batch[1:]:
                if not f1.merge(NPYT(f).load(mmap_mode="r")):
                    return False
            # 改个名字，防止重复合并
            f = f1.filename().with_suffix('.npy_')
            if f1.rename(f):
                logger.info("merge to {} from {}", f, batch)

        return True
